const path = require("node:path");
const fs = require("node:fs");
const crypto = require("node:crypto");
const https = require("node:https");

function checkFileAgainstSHA1(baseDirectory, fileEntry) {
	return new Promise(function (resolve, reject) {
		let filePath = path.join(baseDirectory, fileEntry.name);

		if (!fs.existsSync(filePath)) {
			return resolve({
				"entry": fileEntry,
				"ok": false
			});
		}

		let hash = crypto.createHash("sha1");
		hash.setEncoding("hex");

		let fileInputStream = fs.createReadStream(filePath);

		hash.on("readable", function () {
			let data = hash.read();
			if (data) {
				resolve({
					"entry": fileEntry,
					"ok": data == fileEntry.hash
				});
			}
		});

		fileInputStream.pipe(hash);
	});
}

function downloadFile(url, filePath) {
	return new Promise(function (resolve, reject) {
		let fileOutputStream = fs.createWriteStream(filePath);

		let request = https.request(url, {
			"method": "GET"
		}, function (response) {
			if (response.statusCode !== 200) return reject(new Error("Got a non-OK response while trying to download a file."));

			fileOutputStream.on("end", function () {
				resolve(true);
			});

			response.pipe(fileOutputStream);
		});

		request.on("error", function (error) {
			reject(error);
		});

		request.end();
	});
}

function processDownloadQueue(baseDirectory, baseURL, fileEntries) {
	return new Promise(function (resolve, reject) {
		let runningDownloads = 0;
		let queue = [...fileEntries];

		let finishCallback = function () {
			if (queue.length > 0) {

			}
		};

		for (let d = 0; d < module.exports.concurrentDownloads; d++) {
			if (queue.length < 1) break;

			let currentEntry = queue.shift();

			downloadFile(baseURL + currentEntry.hash, path.join(baseDirectory, currentEntry.name)).then(finishCallback).catch((function (error) {
				if (!this.retries) {
					this.retries = 0;
				}
				if (this.retries > 3) return reject(error);

				this.retries++;
				queue.push(this);
			}).bind(currentEntry));
		}
	});
}

module.exports = {
	"concurrentDownloads": 8,
	"checkFiles": function (manifest, baseDirectory) {
		return new Promise(function(resolve, reject) {
			let checkTasks = [];

			for (let f = 0; f < manifest.files.length; f++) {
				let fileEntry = manifest.files[f];

				checkTasks.push(checkFileAgainstSHA1(baseDirectory, fileEntry));
			}

			Promise.all(checkTasks).then(function (results) {
				resolve(results);
			}).catch(function (error) {
				reject(error);
			});
		});
	},
	"downloadFiles": function (baseDirectory, baseURL, fileEntries) {
		return new Promise(function (resolve, reject) {
			processDownloadQueue(baseDirectory, baseURL, fileEntries.filter(function (entry) {
				return !entry.ok;
			})).then(function () {
				resolve(true);
			}).catch(function (error) {
				reject(error);
			});
		});
	}
};
